/*******************************************************************************
  File: FactoryCollection.h
  Copyright (c) 1999 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, to prevent the 
  continued distribution of this Software under an open source distribution
  model. This includes, but is not limited to, such things as copyrighting
  or claiming authorship of the Software.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
$Header: $
  Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  FactoryCollection is (as the name suggests) a collection of factories.
  It is intended to be used in place of a collection of Prototyes (GoF) 
  to create objects of derived classes in a heirarchy.
  
  Fundamentally, a FactoryCollection provides two services: first it maps 
  type_info objects into user defined name strings. Second, it maps user
  defined name strings into factories that can create objects of that type.
  
//#----------------------------------------------------------------------------|
Usage:
	A FactoryCollection is a template for a mono-state object. 
	First you specialize it with a base class type
		typedef FactoryCollection<BaseType> Factories;
	Then you register the derived types with their appropriate names. Note,
	the base class can be registered as well if it not abstract.
		Factories::register<DerivedType1>("Derived1");
		Factories::register<DerivedType2>("Derived2");
		Factories::register<BaseType>("Base");
	It is an error to try to register the same type or name more than once. 
	This throws an exception (domain_error)
	
	Then you can look up name strings with corresponding type_info information
		std::string type = Factories::find(typeid(DerivedType1)); // returns "Derived1"
	This throws an exception (invalid_argument) if the type_info has not been
	registered (see below for a non-exception version).
		
	Or you can get back a factory object by looking up the name
		Factories::Factory& maker = Factories::find("Derived2");
	The latter is usually used to create a new object of the required type, so it 
	is more commonly used like so
		BaseType* obj = Factories.find("Derived2").make();
	This also throws invalid_argument if the string has not been registered.
	
	For situations where exceptions are probably not acceptable there are the
	following functions:
		template<class T> bool register(const std::string&, std::nothrow_t);
		std::string find(const std::type_info&, std::nothrow_t);
		FactoryCollection<>::Factory* find(const std::string&, std::nothrow_t);
	The first of these returns a true if it suceeded, false otherwise.
	The second returns an empty string if it fails. The third returns a null 
	pointer. 
	
	Note carefully that the nothrow argument only pertains to the
	exception that would be thrown if the function fails in its primary
	operation. Lower level operations (such as std::string copy constructor)
	may throw exceptions which are not prevented.
//%****************************************************************************/
#ifndef UTIL_FACTORY_COLLECTION_H
#define UTIL_FACTORY_COLLECTION_H

// platform
#include <Util/platform.h>

// Standard C++ headers
#include <std/stdexcept>
#include <std/map>
#include <std/string>
#include <std/typeinfo>
#include <new>		// nothrow_t

namespace Util {

export template<typename BaseType>
class FactoryCollection {
public:
	// types
	struct Factory {
		virtual BaseType*	make() const = 0;
	};
	
	template<typename ConcreteType>
	struct TFactory : public Factory {
		virtual BaseType* 	make() const;
	};
	
private:
	// data
	struct TypeInfoComp : std::binary_function<const std::type_info*, 
	                                           const std::type_info*, bool> {
		bool operator()(const std::type_info* lhs, const std::type_info* rhs) const
		{ return lhs->before(*rhs); }
	};
	typedef std::map<const std::type_info*, std::string, TypeInfoComp>	TypeNameMap;
	typedef std::map<std::string, Factory*>								FactoryMap;
	
	static TypeNameMap			_typeNames;
	static FactoryMap			_factories;

public:
	// access functions
	static std::string			find(const std::type_info& type);
	static std::string			find(const std::type_info& type, std::nothrow_t);
	
	static Factory&				find(const std::string& name);
	static Factory*				find(const std::string& name, std::nothrow_t);
	
	template<typename ConcreteClass>
	static void					register_type(const std::string& name);
	template<typename ConcreteClass>
	static bool					register_type(const std::string& name, std::nothrow_t);
};

} //> namespace Util

#ifdef EXPORT_NOT_SUPPORTED
#include "FactoryCollection.hxx"
#endif

#endif
