// my stuff
#include "PerlStream.h"

// C/C++ stuff
#ifdef HPUX_ACC
#  include <iostream/strstream.h>
#  include <iostream/iomanip.h>
#else
#  include <strstream>
#  include <iomanip>
#endif

#include <unistd.h>
#include <fcntl.h>

// Perl stuff
#ifdef HPUX_ACC
extern "C" {
#endif

#include <EXTERN.h>
#include <perl.h>

#ifdef HPUX_ACC
}
#endif

// Define static member data.  We don't use the Perl defines directly because
// we don't want the user to have to include all the Perl headers and totally
// pollute their namespace.
const unsigned int PerlStreamBuf::eVoid          = G_VOID;
const unsigned int PerlStreamBuf::eScalar        = G_SCALAR;
const unsigned int PerlStreamBuf::eArray         = G_ARRAY;

PerlStreamBuf::PerlStreamBuf(unsigned int exception_flags)
throw()
    : streambuf(), mContext(eScalar), mExceptionFlags(exception_flags)
{}

PerlStreamBuf::~PerlStreamBuf()
throw()
{ freePerlValues(); }

// Function: overflow
// Description:
//     This gets called for every character sent to the base class.  Just
//     append it to the stuff we have already buffered.
int
PerlStreamBuf::overflow(int c)
throw()
{
    mBuffer += static_cast<char>(c);
    return c;
}

// Function: xsputn
// Description:
//     Kind of the multi-character version of overflow().  If we don't define
//     it, the base class will call overflow() on each character, so it is
//     just an optimization.
streamsize
PerlStreamBuf::xsputn(const char* s, streamsize nchars)
throw()
{
    mBuffer.append(s, nchars);
    return nchars;
}

// Function: seekoff
// Description:
//     Make the PerlStream look like a real stream with seek capability.  This
//     seeks relative to the beginning, current, or end of the buffer.  Note
//     that all characters after the seek position are erased.  This is not
//     really true seek behavior; we really should have a current write pointer
//     that can be in the middle of the buffer.  But, it complicates things a
//     little and I couldn't imagine seeks being used much in a PerlStream.
streampos
PerlStreamBuf::seekoff(streamoff offset, ios::seek_dir way, int mode)
throw()
{
    if (!(mode & ios::out))
	// This is not an output stream, so error out
	return static_cast<streampos>(-1);

    // Figure out the real seek position.  From the beginning is what we
    // want.  The current position is the same as the end position, so 'cur'
    // and 'end' do the same thing.
    switch (way) {
      case ios::beg:    // Offset relative to beginning of buffer
	break;

      case ios::cur:    // Offset relative to current write position
      case ios::end:    // Offset relative to end of buffer
	offset += mBuffer.length();
	break;

      default:
	return static_cast<streampos>(-1);
    }

    if (offset > static_cast<streamoff>( mBuffer.length() ) || offset < 0) {
	// Out of range
	return static_cast<streampos>(-1);
    } else {
	streampos pos = static_cast<streampos>(offset);
	mBuffer.erase(pos);
	return pos;
    }
}

// Function: seekpos
// Description:
//     Make the PerlStream look like a real stream with seek capability.  This
//     seeks to the given position.  Note that all characters after the seek
//     position are erased.  This is not really true seek behavior; we really
//     should have a current write pointer that can be in the middle of the
//     buffer.  But, it complicates things a little and I couldn't imagine
//     seeks being used much in a PerlStream.
streampos
PerlStreamBuf::seekpos(streampos pos, int mode)
throw()
{
    if (!(mode & ios::out))
	return static_cast<streampos>(-1);

    if (pos > static_cast<streampos>( mBuffer.length() )) {
	return static_cast<streampos>(-1);
    } else {
	mBuffer.erase(pos);
	return pos;
    }
}

// Function: execute
// Description:
//     Execute the Perl code that has been sent to the buffer and get Perl
//     return values.
bool
PerlStreamBuf::execute()
throw(PerlEvalError)
{
    string errstr;
    dSP;
    I32 ax;
    
    // Free return values from the previous eval
    freePerlValues();

    // If nothing is to be sent to Perl, do nothing.
    if (mBuffer.length() == 0)
	return true;

    ENTER;
    SAVETMPS;
    PUSHMARK(sp);

    // Create the string we want to send to Perl, and send it.  Use
    // perl_eval_sv instead of perl_eval_pv because the former takes flags
    // and the latter creates an SV anyway.  The documentation doesn't say
    // that I should undef ERRSV, but I found it to be necessary.
    sv_setsv(ERRSV, &PL_sv_undef);
    PUTBACK;
    SV *code = sv_2mortal( newSVpv( mBuffer.data(), mBuffer.length() ) );
    I32 nstackvals = perl_eval_sv(code, mContext|G_KEEPERR);

    // Now get the response
    SPAGAIN;
    if (SvTRUE(ERRSV)) {
	// Something went wrong
	errstr = "Evaluation failed: '" + mBuffer + "': " + SvPV(ERRSV, PL_na);
	for (I32 i = 0; i < nstackvals; ++i)
	    POPs;
    } else {
	// Get the return values from the Perl stack.  In XS code, the
	// ax/stack_base stuff is wrapped in macros, but no such macros
	// exist outside of XS.
//	sp -= nstackvals;
//	ax = (sp - stack_base) + 1;
	mRtnVals.resize(nstackvals);
	for (I32 i = nstackvals-1; i >= 0; --i) {
	    SV* val = POPs;
	    SvREFCNT_inc( val );
	    mRtnVals[i] = val;
	}
    }
    PUTBACK;
    FREETMPS;
    LEAVE;

    // Initialize things for the next go around
    mBuffer.erase();
    mContext = eScalar;
    if (!errstr.empty()) {
	if (mExceptionFlags & PerlStream::eEvalFail)
	    throw PerlEvalError( errstr );
	return false;
    } else {
	return true;
    }
}

PerlValue
PerlStreamBuf::operator[](int index)
throw(PerlOutOfBounds)
{
    if (index < 0)
	index = mRtnVals.size() + index;
    if (index >= 0 && index < static_cast<int>(mRtnVals.size())) {
	return PerlValue( mRtnVals[index], mExceptionFlags );
    } else if (mExceptionFlags & PerlStream::eArrayOutOfBounds) {
	// Index is out of range, and the user wants us to throw an exception.
	// Use an ostrstream until libraries have the better ostringstream.
	ostrstream buf;
	buf << "Index '" << index <<
		"' of PerlStream array out of bounds: ";
	if (mRtnVals.empty())
	    buf << "array is empty";
	else
	    buf << "valid range = [0," << mRtnVals.size()-1 << "]";
	throw PerlOutOfBounds( buf.str() );
    } else {
	// Index is out of range, but no exception is to be thrown, so just
	// make the return value 'undef'
	return PerlValue( &PL_sv_undef, mExceptionFlags );
    }
}
    
// Function: freePerlValues
// Description:
//    "Free" the Perl return values that we are using by decrementing the
//    reference count of each one.
void
PerlStreamBuf::freePerlValues()
throw()
{
    for (vector<SV*>::iterator p = mRtnVals.begin(); p != mRtnVals.end(); ++p) {
	if (*p) {
	    SvREFCNT_dec(*p);
	    *p = 0;
	}
    }
    mRtnVals.clear();
}
