#ifndef _PERL_STREAM_H
#define _PERL_STREAM_H

// local stuff
#include "PerlValue.h"
#include "PerlExceptions.h"

// STL
#include <string>
#include <vector>

// The streambuf that is used by PerlStream
class PerlStreamBuf : public streambuf {
public:
    // The three types of Perl execution contexts.  We don't use the Perl
    // defines directly because we don't want the user to have to include all
    // the Perl headers and totally pollute their namespace.
    static const unsigned int eVoid;
    static const unsigned int eScalar;
    static const unsigned int eArray;
    
private:
    vector<SV*> mRtnVals;         // Array of values returned from Perl eval
    string mBuffer;               // Buffer to hold code sent to Perl
    unsigned int mContext;        // Execution context for the Perl eval
    unsigned int mExceptionFlags; // A mask of conditions that throw exceptions

public:
    PerlStreamBuf(unsigned int exception_flags) throw();
    ~PerlStreamBuf() throw();

    // Functions from the base class
    virtual int overflow(int c) throw();
    virtual streamsize xsputn(const char* s, streamsize nchars) throw();
    virtual streampos seekoff(streamoff offset, ios::seek_dir dir,
	    int mode = ios::in|ios::out) throw();
    virtual streampos seekpos(streampos pos, int mode = ios::in|ios::out)
	    throw();

    // Execute the Perl code we have collected so far
    bool execute() throw(PerlEvalError);

    // Access Perl return values.  'index' can be negative - index from end.
    PerlValue operator[](int index) throw(PerlOutOfBounds);

    void setContext(unsigned int context) throw()
    { mContext = context; }

    void setExceptionFlags(unsigned int flags) throw()
    { mExceptionFlags = flags; }

    // Get the number of values returned from Perl eval
    size_t nVals() const throw()
    { return mRtnVals.size(); }

private:
    void freePerlValues() throw();
};

// A stream class that acts as a wrapper around the Perl eval function.
// Most of the functions are just wrappers around the real functionality
// in the PerlStreamBuf class.
class PerlStream : public ostream {
protected:
    PerlStreamBuf buf;

public:
    // The various exception flags
    enum {eEvalFail=0x01, eBadCast=0x02, eArrayOutOfBounds=0x04,
	  eHashOutOfBounds=0x08, eAll=~0};

    PerlStream(unsigned int exception_flags=0) throw()
       : ostream(&buf), buf(exception_flags) {}

    void setContext(unsigned int context) throw()
    { buf.setContext(context); }

    void setExceptionFlags(unsigned int flags) throw()
    { buf.setExceptionFlags(flags); }

    PerlStream& scalar() throw()
    { buf.setContext(PerlStreamBuf::eScalar); return *this; }

    PerlStream& array() throw()
    { buf.setContext(PerlStreamBuf::eArray); return *this; }

    PerlStream& voidc() throw()
    { buf.setContext(PerlStreamBuf::eVoid); return *this; }

    PerlStream& execute() throw(PerlEvalError)
    { buf.execute(); return *this; }

    PerlValue operator[](int index) throw(PerlOutOfBounds)
    { return buf[index]; }

    size_t nVals() const throw()
    { return buf.nVals(); }
};

// A function to allow iomanip trickery to work.  I want to be able to do:
//    PerlStream ps;
//    ps << ... << perl_execute;
// Since perl_execute() is defined to take an ostream and return an ostream,
// the iomanip stuff
//    ostream& operator<<(ostream& os, ostream& (*funcptr)(ostream&))
//   { return funcptr(os); }
// gets used to make this happen.
inline ostream&
perl_execute(ostream& o)
{
    PerlStream* pso = dynamic_cast<PerlStream*>(&o);
    if (pso)
	return pso->execute();
    else
	return o;            // do nothing if used on another ostream
}

#endif
