// my stuff
#include "PerlValue.h"
#include "PerlStream.h"

// C/C++
#ifdef HPUX_ACC
#include <iostream/strstream.h>
#else
#include <strstream>
#endif

// Perl
#ifdef HPUX_ACC
extern "C" {
#endif

#include <EXTERN.h>
#include <perl.h>

#ifdef HPUX_ACC
}
#endif

// Function: Constructor for array_iter
// Description:
//    Check that the given PerlValue is indeed an array reference.  If so,
//    get the array length so we know when we are at the end of the array.
PerlValue::array_iter::array_iter(const PerlValue& array_ref)
throw(PerlBadCast)
    : index(0), array_size(0), exception_flags(array_ref.exception_flags)
{
    array = SvROK(array_ref.value) ?
	    reinterpret_cast<AV*>( SvRV(array_ref.value) ) : 0;
    if (!array || SvTYPE(array) != SVt_PVAV) {
	if (array_ref.exception_flags & PerlStream::eBadCast) {
	    string errstr = string("Casting from a Perl ") +
		    array_ref.typeStr() + " to a Perl array reference";
	    throw PerlBadCast(errstr);
	}
    } else {

	// av_len does not really return the length - it returns $#array.
	array_size = av_len(array) + 1;
    }
}

// Function: * operator for array_iter
// Description:
//    Get the SV* that the iterator is pointing at and convert it to a
//    PerlValue.
PerlValue
PerlValue::array_iter::operator*() const
throw(PerlOutOfBounds)
{
    // Check for out of bounds
    if (index >= 0 && index < array_size) {
	SV **rtn = av_fetch( array, index, 0 );
	if (rtn)
	    return PerlValue( *rtn, exception_flags );
    }
    
    // Something when awry - either the array index was out of bounds or Perl
    // returned an error
    if (exception_flags & PerlStream::eArrayOutOfBounds)
	throw PerlOutOfBounds("Perl array iterator out of bounds");
    return PerlValue( &PL_sv_undef, exception_flags );
}

// Function: Constructor for hash_iter
// Description:
//    Check that the given PerlValue is indeed a hash reference.  If so,
//    do the Perl stuff to point to the first element of the hash.
PerlValue::hash_iter::hash_iter(const PerlValue& hash_ref)
throw(PerlBadCast)
    : hash_entry(0), exception_flags(hash_ref.exception_flags)
{
    hash = SvROK(hash_ref.value) ?
	    reinterpret_cast<HV*>( SvRV(hash_ref.value) ) : 0;
    if (!hash || SvTYPE(hash) != SVt_PVHV) {
	if (hash_ref.exception_flags & PerlStream::eBadCast) {
	    string errstr = string("Casting from a Perl ") +
		    hash_ref.typeStr() + " to a Perl hash reference";
	    throw PerlBadCast(errstr);
	}
    } else {
	hv_iterinit(hash);
	hash_entry = hv_iternext(hash);
    }
}

// Function: * operator for hash_iter
// Description:
//    Get the key and SV* that the iterator is pointing at and convert the
//    SV* to a PerlValue.  The hash_entry will be NULL if we reached the
//    end of the hash.
pair<string,PerlValue>
PerlValue::hash_iter::operator*() const
throw(PerlOutOfBounds)
{
    if (hash_entry) {
	I32 key_len;
	const char *key = hv_iterkey(hash_entry, &key_len);
	SV* hash_value = hv_iterval(hash, hash_entry);
	return make_pair( string(key, key_len), PerlValue(hash_value, exception_flags) );
    } else if (exception_flags & PerlStream::eHashOutOfBounds) {
	throw PerlOutOfBounds("Perl hash iterator out of bounds");
    } else {
	return make_pair( string(), PerlValue(&PL_sv_undef, exception_flags) );
    }
}

// Function: increment operator for hash_iter
// Description:
//    Tell Perl to go to the next hash element.
PerlValue::hash_iter&
PerlValue::hash_iter::operator++()
throw()
{
    hash_entry = hv_iternext(hash);
    return *this;
}

// Function: is*()
// Description:
//    Ask Perl about the SV*.
bool
PerlValue::isDefined() const
throw()
{ return SvOK(value); }

bool
PerlValue::isInt() const
throw()
{ return SvIOK(value); }

bool
PerlValue::isDouble() const
throw()
{ return SvNOK(value); }

bool
PerlValue::isNumber() const
throw()
{ return SvNIOK(value); }

bool
PerlValue::isString() const
throw()
{ return SvPOK(value); }

bool
PerlValue::isScalarRef() const
throw()
{
    if (SvROK(value)) {
	int type = SvTYPE( SvRV(value) );
	if (type == SVt_IV || type == SVt_NV || type == SVt_PV ||
		type == SVt_RV)
	    return true;
    }
    return false;
}

bool
PerlValue::isArrayRef() const
throw()
{ return SvROK(value) && SvTYPE(SvRV(value)) == SVt_PVAV; }

bool
PerlValue::isHashRef() const
throw()
{ return SvROK(value) && SvTYPE(SvRV(value)) == SVt_PVHV; }

// Function: conversion operators
// Description:
//    Convert the SV* to a C type.  Check to see if it really is that type
//    before asking Perl to do the conversion.  If the type is bad, either
//    throw an exception or return an innocuous value.
int
PerlValue::Int() const
throw(PerlBadCast)
{
    if (SvIOK(value) || SvNOK(value) || SvPOK(value)) {
	return SvIV(value);
    } else if (exception_flags & PerlStream::eBadCast) {
	string errstr = string("Casting from a Perl ") + typeStr() +
		" to an int";
	throw PerlBadCast(errstr);
    }
    return 0;         // make it a 0
}

double
PerlValue::Double() const
throw(PerlBadCast)
{
    if (SvIOK(value) || SvNOK(value) || SvPOK(value)) {
	return SvNV(value);
    } else if (exception_flags & PerlStream::eBadCast) {
	string errstr = string("Casting from a Perl ") + typeStr() +
		" to a double";
	throw PerlBadCast(errstr);
    }
    return 0.0;
}

string
PerlValue::String() const
throw(PerlBadCast)
{ 
    if (SvIOK(value) || SvNOK(value) || SvPOK(value)) {
	size_t len;
	const char *str = SvPV(value, len);
	return string(str, len);
    } else if (exception_flags & PerlStream::eBadCast) {
	string errstr = string("Casting from a Perl ") + typeStr() +
		" to a string";
	throw PerlBadCast(errstr);
    }
    return string();
}

bool
PerlValue::Bool() const
throw()
{
    return SvTRUE(value);
}
    
// Function: typeStr()
// Description:
//    Return a string describing the type of the SV*.  This is mainly used
//    for error messages.
const char*
PerlValue::typeStr() const
throw()
{
    switch (SvTYPE(value)) {
      case SVt_NULL: return "undef";
      case SVt_IV:   return "int";
      case SVt_NV:   return "double";
      case SVt_PV:   return "string";
      case SVt_PVAV: return "array";
      case SVt_PVHV: return "hash";
      case SVt_PVCV: return "code";
      case SVt_PVGV: return "glob";
      case SVt_PVMG: return "blessed scalar";
      case SVt_RV:
	switch (SvTYPE( SvRV(value) )) {
	  case SVt_IV:
	  case SVt_NV:
	  case SVt_PV:
	  case SVt_RV: return "scalar reference";
	  case SVt_PVAV: return "array reference";
	  case SVt_PVHV: return "hash reference";
	  case SVt_PVCV: return "code reference";
	  case SVt_PVGV: return "glob reference";
	  case SVt_PVMG: return "blessed reference";
	  default: return "unknown reference";
	}
      default: return "unknown type";
    }
}

// Function: * operator for PerlValue
// Description:
//    Dereference, Perl style, the SV* after checking that the SV* is indeed
//    a scalar reference.
PerlValue
PerlValue::operator*() const
throw(PerlBadCast)
{
    if (isScalarRef()) {
	return PerlValue( SvRV(value), exception_flags );
    } else if (exception_flags & PerlStream::eBadCast) {
	string errstr = string("Casting from a Perl ") + typeStr() +
		" to a Perl reference";
	throw PerlBadCast(errstr);
    }
    return PerlValue( &PL_sv_undef, exception_flags );
}

// Function: at
// Description:
//    Check that the SV* is an array reference, and, if so, get the SV* that
//    is at the given index in the array.  If the index is negative, index
//    into the array from the end, just like Perl.
PerlValue
PerlValue::at( int index ) const
throw(PerlBadCast, PerlOutOfBounds)
{
    if (!isArrayRef()) {
	if (exception_flags & PerlStream::eBadCast) {
	    string errstr = string("Casting from a Perl ") + typeStr() +
		    " to a Perl array reference";
	    throw PerlBadCast(errstr);
	}
	return PerlValue( &PL_sv_undef, exception_flags );
    }

    AV* array = reinterpret_cast<AV*>( SvRV(value) );

    // Do negative indexing like Perl - from the end of the array.  av_len does
    // not really return the length - it returns $#array.
    int original_index = index;
    if (index < 0)
	index = av_len(array) + 1 + index;

    if (index >= 0 && index <= av_len(array)) {
	SV **rtn = av_fetch( array, index, 0 );
	if (rtn)
	    return PerlValue( *rtn, exception_flags );
    }
    
    // Something when awry - either the array index was out of bounds or Perl
    // returned an error
    if (exception_flags & PerlStream::eArrayOutOfBounds) {
	ostrstream buf;
	buf << "Index '" << original_index <<
		"' of Perl array out of bounds: valid range = [" <<
		-av_len(array)-1 << "," << av_len(array) << "]" << '\0';
	throw PerlOutOfBounds(buf.str());
    }
    return PerlValue( &PL_sv_undef, exception_flags );
}

// Function: at
// Description:
//    Check that the SV* is a hash reference, and, if so, lookup the given
//    key in the hash and get the SV* there.
PerlValue
PerlValue::at( const char* hash_key, size_t hash_key_len ) const
throw(PerlBadCast, PerlOutOfBounds)
{
    if (!isHashRef()) {
	if (exception_flags & PerlStream::eBadCast) {
	    string errstr = string("Casting from a Perl ") + typeStr() +
		    " to a Perl hash reference";
	    throw PerlBadCast(errstr);
	}
	return PerlValue( &PL_sv_undef, exception_flags );
    }

    HV* hash = reinterpret_cast<HV*>( SvRV(value) );

    // Do the fetch.  The last parameter tells Perl to not create a new entry
    // if the entry doesn't exist.
    SV **rtn = hv_fetch( hash, hash_key, hash_key_len, 0 );
    if (rtn) {
	return PerlValue( *rtn, exception_flags );
    } else if (exception_flags & PerlStream::eHashOutOfBounds) {
	string errstr = string("Perl hash key '") + hash_key +
		"' does not exist";
	throw PerlOutOfBounds(errstr);
    } else
	return PerlValue( &PL_sv_undef, exception_flags );
}

// Function: size
// Description:
//    If the SV* is an array reference or hash reference, get the number of
//    elements in the data structure.
int
PerlValue::size() const
throw(PerlBadCast)
{
    if (SvROK(value)) {
	SV* unref_value = SvRV(value);
	switch (SvTYPE(unref_value)) {
	  case SVt_PVAV:
	    // av_len does not really return the length - it returns $#array.
	    return av_len( reinterpret_cast<AV*>(unref_value) ) + 1;
	    
	  case SVt_PVHV:
	    return HvKEYS( reinterpret_cast<HV*>(unref_value) );
	}
    }

    // Not an array reference or hash reference
    if (exception_flags & PerlStream::eBadCast) {
	string errstr = string("Casting from a Perl ") + typeStr() +
		" to a Perl array reference";
	throw PerlBadCast(errstr);
    }
    return 0;
}

// Function: << operator for a PerlValue
// Description:
//    Figure out the type of the SV* and print out the appropriate converted
//    value.
ostream&
operator<<(ostream& os, PerlValue val)
throw()
{
    switch (SvTYPE(val.value)) {
      case SVt_NULL: os << "<undef>";                      break;
      case SVt_PV:   os << static_cast<string>( val );     break;
      case SVt_NV:   os << static_cast<double>( val );     break;
      case SVt_IV:   os << static_cast<int>( val );        break;
      case SVt_RV:   os << "REF(" << val.value << ")";     break;
      case SVt_PVAV: os << "ARRAY(" << val.value << ")";   break;
      case SVt_PVHV: os << "HASH(" << val.value << ")";    break;
      case SVt_PVCV: os << "CODE(" << val.value << ")";    break;
      case SVt_PVGV: os << "GLOB(" << val.value << ")";    break;
      case SVt_PVMG: os << "BLESSED(" << val.value << ")"; break;
      default:       os << "UNKNOWN(" << val.value << ")"; break;
    }
    return os;
}
